"""
    Based on Enviroment class in https://github.com/hongzimao/pensieve
"""

import numpy as np

MILLISECONDS_IN_SECOND = 1000.0
B_IN_MB = 1000000.0
BITS_IN_BYTE = 8.0
BITRATE_LEVELS = 7
PACKET_PAYLOAD_PORTION = 0.95
LINK_RTT = 80  # millisec
NOISE_LOW = 0.9
NOISE_HIGH = 1.1
VIDEO_SIZE_FILE = './video_size_'

class Environment:
    def __init__(self, all_cooked_time, all_cooked_bw, random_seed):
        assert len(all_cooked_time) == len(all_cooked_bw)

        np.random.seed(random_seed)
        
        self.all_cooked_time = all_cooked_time
        self.all_cooked_bw = all_cooked_bw
        
        # pick a random trace file
        self.trace_idx = np.random.randint(len(self.all_cooked_time))
        self.cooked_time = self.all_cooked_time[self.trace_idx]
        self.cooked_bw = self.all_cooked_bw[self.trace_idx]
        
        # randomize the start point of the trace
        self.mahimahi_ptr = np.random.randint(1, len(self.cooked_bw))
        self.last_mahimahi_time = self.cooked_time[self.mahimahi_ptr - 1]
        
        # Get size of every segment for all the qualities
        self.video_size = {} # in bytes
        for bitrate in range(1,BITRATE_LEVELS):
            self.video_size[bitrate] = []
            with open(VIDEO_SIZE_FILE + str(bitrate)) as f:
                for line in f:
                    self.video_size[bitrate].append(int(line.split()[0]))
    
    def get_video_chunk(self, quality, video_chunk_counter):
        assert quality > 0
        assert quality < BITRATE_LEVELS
        
        # Size of selected segment
        video_chunk_size = self.video_size[quality][video_chunk_counter]
        
        delay = 0.0
        video_chunk_counter_sent = 0
        
        # Download segment
        while True:
            # Get network parameters
            throughput = self.cooked_bw[self.mahimahi_ptr] * B_IN_MB / BITS_IN_BYTE
            duration = self.cooked_time[self.mahimahi_ptr] - self.last_mahimahi_time
            
            packet_payload = throughput * duration * PACKET_PAYLOAD_PORTION
            
            # Check if reamining video does not need whole packet to be sent
            if video_chunk_counter_sent + packet_payload > video_chunk_size:
                
                fractional_time = (video_chunk_size - video_chunk_counter_sent) / throughput / PACKET_PAYLOAD_PORTION
                delay += fractional_time
                self.last_mahimahi_time += fractional_time
                assert(self.last_mahimahi_time <= self.cooked_time[self.mahimahi_ptr])
                break
            
            video_chunk_counter_sent += packet_payload
            delay += duration
            self.last_mahimahi_time = self.cooked_time[self.mahimahi_ptr]
            self.mahimahi_ptr += 1
            
            if self.mahimahi_ptr >= len(self.cooked_bw):
                # loop back in the beginning
                # note: trace file starts with time 0
                self.mahimahi_ptr = 1
                self.last_mahimahi_time = 0
        
        # Add RTT to the download time
        delay *= MILLISECONDS_IN_SECOND
        rtt = LINK_RTT * np.random.uniform(NOISE_LOW, NOISE_HIGH)
        delay += rtt
        
        return delay, video_chunk_size, rtt
