import env
import load_trace
import abr
from datetime import datetime
import numpy as np
from IPython import display
from tabulate import tabulate

VIDEO_CHUNCK_LEN = 4000.0  # in millisec
TOTAL_VIDEO_CHUNCK = 48


# Kbps
AVAILABLE_QUALITY_LEVELS = {
    0 : 0,
    1 : 300,
    2 : 750,
    3 : 1200,
    4 : 1850,
    5 : 2850,
    6 : 4300
}

BITS_IN_BYTE = 8.0
TRACES = './cooked_traces/'
MS_IN_S = 1000.0
DELTA = 0
REBUFFER_PENALTY = 4.3
MBPS_IN_KBPS = 0.001

all_cooked_time, all_cooked_bw, _ = load_trace.load_trace(TRACES)

class PlaybackResult:
    def __init__(self, name, bitrates, buffer_levels, times, qualities, rebuffering_time, wasted_time, positive_switches, negative_switches, qoe):
        self.name = name
        
        for i in range(len(qualities)):
            if qualities[i] == 0 and i != 0:
                qualities[i] = qualities[i-1]
        
        self.qualities = qualities
        self.bitrates = bitrates
        self.average_bitrate = round(sum(self.bitrates) / len(self.bitrates), 2)
        
        self.buffer_levels = [x / 1000 for x in buffer_levels]
        self.average_buffer_level = round(sum(self.buffer_levels) / len(self.buffer_levels), 2)
        
        self.times = [x / 1000 for x in times]
        self.rebuffering_time = round(rebuffering_time / 1000, 2)
        self.wasted_time = round(wasted_time / 1000, 2)
        
        self.positive_switches = positive_switches
        self.negative_switches = negative_switches
        
        self.qoe = qoe
        
    
    def to_row(self):
        return [self.name.replace("ABR",""),self.average_buffer_level, self.average_bitrate, self.rebuffering_time, self.wasted_time, self.positive_switches, self.negative_switches, self.qoe]

class Player:
    def __init__(self, abr_algorithm, buffer_size, buffer_playback_start_threshold, random_seed):
        self.buffer_playback_start_threshold = buffer_playback_start_threshold
        self.buffer_size = buffer_size
        self.abr_algorithm = abr_algorithm
        self.initial_delay = 0
        self.rebuffering = 0
        self.total_wasted = 0
        self.buffer_level = 0
        self.quality = 1
        self.chunk_counter = 0
        self.net_env = env.Environment(all_cooked_time, all_cooked_bw, random_seed)
        self.playing = False
        self.time = 0
        self.positive_switches = 0
        self.negative_switches = 0
        self.qoe = 0
        
    def playback(self):
        qualities = [1]
        times = [0]
        buffer_levels = [0]
        rebuffering_event = 0
        while True:
            # Download segment
            download_time, chunk_size, rtt = self.net_env.get_video_chunk(self.quality, self.chunk_counter)
            self.time += download_time

            # Update the video buffer by the elapsed time
            if self.playing:
                self.buffer_level = self.buffer_level - download_time

            # Did the buffer run empty?
            if(self.buffer_level < 0):
                if(self.chunk_counter == 1):
                    self.initial_delay = -(self.buffer_level)
                else:
                    rebuffering_event = -(self.buffer_level)
                    self.rebuffering += rebuffering_event
                self.buffer_level = 0
                # Update statistics
                times.append(self.time - rebuffering_event)
                buffer_levels.append(self.buffer_level)
                qualities.append(self.quality)

            # Update statistics
            times.append(self.time)
            buffer_levels.append(self.buffer_level)
            qualities.append(self.quality)
            
            # Add downloaded segment to the buffer
            self.buffer_level += VIDEO_CHUNCK_LEN
            
            # Less than one chunk space left? Then we have to wait!
            # Was time wasted without downloading anything?
            if(self.buffer_level > self.buffer_size - VIDEO_CHUNCK_LEN):
                times.append(self.time + DELTA)
                buffer_levels.append(self.buffer_level)
                qualities.append(0)

                wasted = (self.buffer_level - (self.buffer_size - VIDEO_CHUNCK_LEN))
                self.total_wasted += wasted
                self.buffer_level = self.buffer_size - VIDEO_CHUNCK_LEN
                
                # Update statistics
                self.time += wasted
                
                times.append(self.time)
                buffer_levels.append(self.buffer_level)
                qualities.append(0)

            # If enough video in the buffer, start playing
            if not self.playing & (self.buffer_level >= self.buffer_playback_start_threshold):
                self.playing = True
            self.chunk_counter += 1

            # Get next quality
            throughput = (float(chunk_size * BITS_IN_BYTE)) / float(download_time)
            self.quality = self.abr_algorithm.get_next_quality_level(self.buffer_level, rtt, throughput)

            #Update statistics
            times.append(self.time + DELTA)
            buffer_levels.append(self.buffer_level)
            qualities.append(self.quality)
            
            
            # If all the segments downloaded, stop the playback
            if(self.chunk_counter > TOTAL_VIDEO_CHUNCK):
                break
        
        # Count the number of quality switches
        # Don't consider wasted time
        downloaded_qualities = [i for i in qualities if i != 0]
        self.positive_switches = np.sum([1 for i in (downloaded_qualities - np.roll(downloaded_qualities,-1))[:-1] if i < 0])
        self.negative_switches = np.sum([1 for i in (downloaded_qualities - np.roll(downloaded_qualities,-1))[:-1] if i > 0])
        
        
        # Qualities to bitrates
        bitrates = [AVAILABLE_QUALITY_LEVELS[i] * MBPS_IN_KBPS for i in qualities]
        
        # Compute penalty due to bitrate switches
        quality_switches_penalty = np.sum((bitrates - np.roll(bitrates,-1))[:-1])
        
        # QoE metric
        self.qoe = round(np.sum(bitrates) / len(bitrates) - REBUFFER_PENALTY * self.rebuffering - np.sum(quality_switches_penalty), 2)

        return PlaybackResult(self.abr_algorithm.__class__.__name__, bitrates, buffer_levels, times, qualities, self.rebuffering, self.total_wasted, self.positive_switches, self.negative_switches, self.qoe)
    
    
def display_quality_levels(AVAILABLE_QUALITY_LEVELS):
    v= [["Bitrate"] + list(AVAILABLE_QUALITY_LEVELS.values())]
    h= ["Quality Index"] + list(AVAILABLE_QUALITY_LEVELS.keys())
    return display.HTML(tabulate(v, headers=h,tablefmt="html"))